import argparse
import json
import sys

from .config import build_provider_from_file
from .constants import DEBIAN_VERSIONS
from .exceptions import UserError
from .minions import (
    create_minion,
    destroy_minion,
)
from .master import create_master
from .providers import provider_map
from .roles import get_minion_arguments_for_role, get_provider_for_role
from .version import __version__

class TerminalColors:
    WARNING = '\033[33m'
    FAIL = '\033[91m'
    RESET = '\033[0m'


def main(argv=None):
    try:
        cli = HartCLI()
        args = cli.get_args(argv)
        args.action(args)
    except UserError as error:
        log_error(str(error))
        sys.exit(1)


def log_warning(message, end='\n'):
    log_to_stderr_with_color(message, TerminalColors.WARNING, end)


def log_error(message, end='\n'):
    log_to_stderr_with_color(message, TerminalColors.FAIL, end)


def log_to_stderr_with_color(message, color, end):
    out = sys.stderr
    if out.isatty():
        out.write(color)
        out.write(message)
        out.write(TerminalColors.RESET)
    else:
        out.write(message)
    if end:
        out.write(end)
    out.flush()


class HartCLI:
    def __init__(self):
        if sys.getfilesystemencoding() == 'ascii':
            raise UserError('Your system has incorrect locale settings, '
                'leading to non-unicode default IO. Set f. ex '
                'LC_CTYPE=en_US.UTF-8 and PYTHONIOENCODING=utf-8 to fix this.')


    def get_args(self, argv):
        parser = argparse.ArgumentParser(prog='hart', add_help=False)

        parser.add_argument('-P', '--provider', choices=provider_map.keys(),
            help='Which VPS provider to use.')
        parser.add_argument('-R', '--region',
            help='Which region to create the node in. Default: %(default)s')
        parser.add_argument('-c', '--config', default='/etc/hart.toml',
            help='Path to config file with credentails. Default: %(default)s')
        # Explicitly add help to be able to parse the provider before printing the help
        parser.add_argument('-h', '--help', action='store_true', help='Print help')
        parser.add_argument('-v', '--version', action='version', version='hart v%s' % __version__)

        subparsers = parser.add_subparsers(dest='command',
            title='Commands',
            help='What do you want to do?')

        create_minion_from_role_parser = self.add_create_minion_from_role_parser(subparsers)
        create_minion_parser = self.add_create_minion_parser(subparsers)
        create_master_parser = self.add_create_master_parser(subparsers)
        destroy_minion_parser = self.add_destroy_minion_parser(subparsers)
        list_regions_parser = self.add_list_regions_parser(subparsers)
        list_sizes_parser = self.add_list_sizes_parser(subparsers)

        # Do an initial parse of just the provider arguments, to be able to add
        # provider-specific arguments to the full parse. If a provider is given
        # on the command line that takes precedence, otherwise when creating
        # from a role there might be a provider specified in the config file,
        # use that.
        provider_args, _ = parser.parse_known_args(argv)
        provider = None

        try:
            if provider_args.provider:
                provider = get_provider(provider_args.provider, provider_args.config,
                    provider_args.region)
            elif provider_args.command == 'create-minion-from-role':
                provider = get_provider_for_role(
                    provider_args.config, provider_args.role, provider_args.region)
            else:
                raise UserError('No provider specified')

        except (FileNotFoundError, KeyError):
            # Enable running help without having a valid provider config
            # FileNotFoundError if config is missing entirely, KeyError if the
            # defualt or given provider is missing
            # TODO: This ignores any action that might be given, the defualt
            # parameters for those should be included
            if provider_args.help:
                log_warning('Unable to instantiate provider, add a valid provider config to see '
                    'all available parameters')
                parser.print_help()
                sys.exit(0)
            raise

        # Add the same arguments to create-minion-from-role as create-minion
        provider.add_create_minion_arguments(create_minion_from_role_parser)
        provider.add_create_minion_arguments(create_minion_parser)
        provider.add_create_minion_arguments(create_master_parser)
        provider.add_destroy_minion_arguments(destroy_minion_parser)
        provider.add_list_regions_arguments(list_regions_parser)
        provider.add_list_sizes_arguments(list_sizes_parser)

        args = parser.parse_args(argv)
        args.provider = provider

        if args.help:
            parser.print_help()
            sys.exit(0)

        if not getattr(args, 'action', None):
            parser.print_help()
            sys.exit(1)

        return args


    def add_create_minion_from_role_parser(self, subparsers):
        parser = subparsers.add_parser('create-minion-from-role', help='Create a new minion with a given role')
        parser.add_argument('role', help='Name of the role')
        self._add_minion_master_shared_arguments(parser)
        parser.set_defaults(action=self.cli_create_minion_from_role)
        return parser


    def add_create_minion_parser(self, subparsers):
        parser = subparsers.add_parser('create-minion', help='Create a new minion')
        self._add_minion_master_shared_arguments(parser)
        parser.set_defaults(action=self.cli_create_minion)
        return parser


    def add_create_master_parser(self, subparsers):
        parser = subparsers.add_parser('create-master', help='Create a new saltmaster')
        self._add_minion_master_shared_arguments(parser)

        parser.add_argument('-a', '--authorize-key',
            help='An ssh public key to add to .ssh/authorized_keys.')
        parser.add_argument('-g', '--grains', type=type_json,
            help="Grains to write to /etc/salt/minion.d/grains.conf")

        parser.set_defaults(action=self.cli_create_master)
        return parser


    def _add_minion_master_shared_arguments(self, parser): # pylint disable=no-self-use
        def type_csv(clistring):
            return clistring.split(',')

        parser.add_argument('minion_id')
        parser.add_argument('-s', '--size',
            help='The size of the node to create. Default varies with provider.')
        parser.add_argument('-t', '--tags', type=type_csv, default=[],
            help='Tags to add to the new node, comma-separated.')
        parser.add_argument('-S', '--script', help='Path to a script to run on '
            'the master after salt has been installed. Use this to customize '
            'your setup and deploy any secrets you might need.')
        parser.add_argument('-p', '--private-networking', action='store_true',
            help='Whether to enable private networking on the node')
        parser.add_argument('-d', '--debian-codename',
            choices=DEBIAN_VERSIONS.keys(), default='buster',
            help='Which debian version to create. Default: %(default)s')
        parser.add_argument('--use-py2', action='store_true',
            help='Use py2 instead of py3 for saltstack.')
        parser.add_argument('--salt-branch', default='latest',
            help='The salt branch to use. Default: %(default)s')
        parser.add_argument('--minion-config', type=type_json,
            help='Minion config in JSON')


    def add_destroy_minion_parser(self, subparsers):
        parser = subparsers.add_parser('destroy-minion', help='Destroy a minion')
        parser.add_argument('minion_id')

        parser.set_defaults(action=self.cli_destroy_minion)
        return parser


    def add_list_regions_parser(self, subparsers):
        parser = subparsers.add_parser('list-regions',
            help='List available regions for a provider')

        parser.set_defaults(action=self.cli_list_regions)
        return parser


    def add_list_sizes_parser(self, subparsers):
        parser = subparsers.add_parser('list-sizes',
            help='List available sizes in a region for a provider')

        parser.set_defaults(action=self.cli_list_sizes)
        return parser


    def cli_create_minion_from_role(self, args):
        kwargs = get_minion_arguments_for_role(
            args.config, args.role, args.provider, args.region, vars(args))
        for key, val in kwargs.items():
            setattr(args, key, val)
        self.cli_create_minion(args)


    def cli_create_minion(self, args):
        kwargs = vars(args)
        try:
            create_minion(**kwargs)
        except KeyboardInterrupt:
            print('Aborted by Ctrl-C or SIGINT, stopping')


    def cli_create_master(self, args):
        kwargs = vars(args)
        try:
            create_master(**kwargs)
        except KeyboardInterrupt:
            print('Aborted by Ctrl-C or SIGINT, stopping')


    def cli_destroy_minion(self, args):
        kwargs = vars(args)
        provider = kwargs.pop('provider')
        minion_id = kwargs.pop('minion_id')
        destroy_minion(minion_id, provider, **kwargs)


    def cli_list_sizes(self, args):
        kwargs = vars(args)
        provider = kwargs.pop('provider')
        for size in provider.get_sizes(**kwargs):
            formatted_memory = '%d' % size.memory if size.memory >= 1 else '%.1f' % size.memory
            print('%d vCPUs, %s GB RAM, %s (%s, $%d/month) %s' % (
                size.cpu,
                formatted_memory,
                size.disk,
                size.id,
                size.monthly_cost,
                size.extras or ''),
            )


    def cli_list_regions(self, args):
        kwargs = vars(args)
        provider = kwargs.pop('provider')
        for location in provider.get_regions(**kwargs):
            print('%s (%s)' % (location.name, location.id))


def type_json(value):
    return json.loads(value)


def get_clean_kwargs_from_args(args):
    kwargs = vars(args)
    del kwargs['provider']
    return kwargs


def get_provider(provider_alias, config_path, region):
    return build_provider_from_file(provider_alias, config_path, region=region)


if __name__ == '__main__':
    main(sys.argv)



from libcloud.compute.base import NodeAuthSSHKey

from .base import BaseProvider, Region
from ..exceptions import UserError


class BaseLibcloudProvider(BaseProvider):
    driver = None

    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        '''Return a tuple of (remote_key, auth_key)'''
        remote_key = self.driver.create_key_pair(key_name, public_key)
        auth_key = NodeAuthSSHKey(remote_key.public_key)
        return remote_key, auth_key


    def destroy_remote_ssh_key(self, remote_key):
        self.driver.delete_key_pair(remote_key)


    def get_size(self, size_name):
        sizes = self.driver.list_sizes()
        for size in sizes:
            if size_name in (size.id, size.name):
                return size

        raise UserError('Unknown size: %s' % size_name)


    def get_location(self, location_id):
        for location in self.driver.list_locations():
            if location_id in (location.name, location.id):
                return location

        raise UserError('Location %s not found' % location_id)


    def destroy_node(self, node, extra=None, **kwargs):
        self.driver.destroy_node(node)


    def get_node(self, node_id):
        if not isinstance(node_id, str):
            node_id = node_id.name
        for node in self.driver.list_nodes():
            if node_id in (node.id, node.name):
                return node

        raise UserError('No node with id %s found in provider %s' % (
            node_id, self.__class__.__name__))


    def get_regions(self, **kwargs):
        regions = []
        for location in self.driver.list_locations():
            regions.append(Region(location.id, location.name))
        regions.sort(key=lambda r: r.name)
        return regions



import base64
import contextlib
import hashlib
import os
import time
import select
import socket

import paramiko
from paramiko.ssh_exception import SSHException


class IgnorePolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        # key.__str__() isn't valid according to str() since it returns bytes, thus
        # calling it directly
        key_digest = hashlib.sha256(key.__str__()).digest()
        fingerprint = 'SHA256:%s' % base64.b64encode(key_digest).decode('utf-8')
        print('Accepting host key type %s with fingerprint %s' % (
            key.get_name(), fingerprint))


def ssh_run_command(client, command, timeout=3, log_stdout=True):
    # Work around circular import
    from .__main__ import log_error

    captured_stdout = []
    session = client.get_transport().open_session()
    session.exec_command(command)
    chunksize = 1024
    start_time = time.time()
    while True:
        (reads_ready, _, _) = select.select([session], [], [], 1)
        if timeout and time.time() - start_time > timeout:
            raise ValueError('Timed out waiting for command %r to finish' % command)

        if not reads_ready:
            continue

        got_data = False
        if session.recv_ready():
            chunk = reads_ready[0].recv(chunksize).decode('utf-8')
            if chunk:
                got_data = True
                captured_stdout.append(chunk)
                if log_stdout:
                    print(chunk, end='')

        if session.recv_stderr_ready():
            chunk = reads_ready[0].recv_stderr(chunksize).decode('utf-8')
            if chunk:
                got_data = True
                log_error(chunk, end='')

        if not got_data:
            break

    while not session.exit_status_ready():
        if timeout and time.time() - start_time > timeout:
            raise ValueError('Timed out waiting for command %r to return an exit code' % command)
        time.sleep(1)

    exit_status = session.recv_exit_status()
    if exit_status != 0:
        raise ValueError('Command %r failed with exit code %d' % (command, exit_status))

    return ''.join(captured_stdout)


@contextlib.contextmanager
def get_verified_ssh_client(ip, ssh_key, canary, username='root'):
    client = connect_to_droplet(ip, ssh_key, username)
    print('Connected')

    # We might not be connected to the right box yet, but we should help seed
    # the random pool as early as possible in the boot sequence. There's nothing
    # sensitive here as we'll disconnect if the canary fails in the next step
    # and the minion will be destroyed. Doing this in addition to seeding over
    # cloud-init since the contents of cloud-init is rarely safe from someone
    # that manages to compromise the server.
    print('Seeding random pool')
    seed_client_random_pool(client)

    # Verify the ssh canary as the first thing to not run any potentially
    # dangerous operations on an untrusted box
    wait_for_verified_ssh_canary(client, canary, should_sudo=username != 'root')
    print('Verified connection')
    try:
        yield client
    except:
        client.close()
        raise


def log_action(action, start_time):
    print('action=%s time=%.2fs' % (action, time.time() - start_time))


def connect_to_droplet(ip, client_ssh_key, username):
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(IgnorePolicy())
    timeout = 120
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            client.connect(ip, username=username, pkey=client_ssh_key, timeout=3)
            log_action('connect', start_time)
            break
        except (socket.error, SSHException) as error:
            print('Could not connect yet, waiting (%s)' % error)
            time.sleep(2)
    else:
        raise ValueError('Failed to connect to new node')

    return client


def seed_client_random_pool(client):
    seed = os.urandom(32)
    stdin, _, _ = client.exec_command('dd of=/dev/random', timeout=3)
    stdin.write(seed)
    stdin.channel.close()


def wait_for_verified_ssh_canary(client, ssh_canary, should_sudo):
    timeout = 15
    start_time = time.time()
    while time.time() - start_time < timeout:
        # The remove is just a matter of cleanup, the canary isn't sensitive
        _, stdout, stderr = client.exec_command(
            'cat /tmp/ssh-canary && {0}rm /tmp/ssh-canary'.format('sudo ' if should_sudo else ''),
            timeout=3)
        if stderr.channel.recv_exit_status() != 0:
            print('No ssh canary yet, waiting (%s)' % ''.join(stderr).strip())
            time.sleep(1)
            continue

        found_canary = ''.join(stdout).strip()
        if found_canary != ssh_canary:
            raise ValueError('ssh canary check failed!')

        log_action('canary-check', start_time)
        break
    else:
        raise ValueError('No canary found before timeout')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import os
import re
from setuptools import setup, find_packages

install_requires = [
    'apache-libcloud>=2.6.1',
    'jinja2',
    'paramiko',
    'pyyaml',
    'toml',
    'ifaddr',
    'boto3',
]

version_file = os.path.join(os.path.dirname(__file__), 'hart', 'version.py')
with open(version_file) as fh:
    version_file_contents = fh.read().strip()
    version_match = re.match(r"__version__ = '(\d+\.\d+.\d+.*)'", version_file_contents)
    version = version_match.group(1)

setup(
    name='hart',
    version=version,
    author='Tarjei Husøy',
    author_email='git@thusoy.com',
    url='https://github.com/thusoy/hart',
    description="Opinionated salt-cloud alternative",
    packages=find_packages(),
    install_requires=install_requires,
    package_data={
        'hart': ['cloud-init/*.sh'],
    },
    entry_points={
        'console_scripts': [
            'hart = hart.__main__:main',
        ]
    },
    classifiers=[
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Web Environment',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        # 'Programming Language :: Python :: 2.6',
        # 'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import base64
import datetime
import os
from collections import namedtuple

import jinja2

from .constants import DEBIAN_VERSIONS
from .exceptions import UserError


HartNode = namedtuple('HartNode', 'minion_id public_ip node provider ssh_key ssh_canary node_extra')


def create_token():
    return base64.urlsafe_b64encode(os.urandom(32)).rstrip(b'=').decode('utf-8')


def get_cloud_init_template(template_name='minion.sh'):
    template_directory = os.path.join(os.path.dirname(__file__), 'cloud-init')
    environment = jinja2.Environment(loader=jinja2.FileSystemLoader(template_directory))
    return environment.get_template(template_name)


def get_saltstack_repo_url(debian_codename, salt_branch, use_py2):
    debian_version = DEBIAN_VERSIONS[debian_codename]
    if use_py2 and debian_version > 9:
        raise UserError('saltstack py2 is only available for debian stretch and older')
    if not use_py2 and debian_version < 9:
        raise UserError('saltstack py3 is only available for debian stretch and newer')
    return 'https://repo.saltstack.com/%s/debian/%s/amd64/%s %s main' % (
        'apt' if use_py2 else 'py3', debian_version, salt_branch, debian_codename)


def build_ssh_key_name(minion_id):
    current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
    return 'hart-temp-for-%s-at-%s' % (minion_id, current_date)
